"""
Useful enum implementations
"""
import json
from string import Template
from typing import List

from aenum import MultiValueEnum


class BaseEOGrowEnum(MultiValueEnum):
    """A base class for `eo-grow` enum classes, implementing common properties and methods"""

    @property
    def id(self) -> int:
        """Returns an ID of an enum type

        :return: An ID
        """
        return self.values[1]

    @property
    def nice_name(self) -> str:
        """Returns nice name of an enum type

        :returns: Nice name
        """
        return self.value.capitalize()

    @property
    def color(self) -> str:
        """Returns class color

        :return: A color in hexadecimal representation
        """
        return self.values[2]

    @property
    def rgb(self) -> List[float]:
        """Returns class color in RGB representation, useful for building matplotlib colormaps

        :return: A color as [R, G, B] list of floats
        """
        return [c / 255.0 for c in self.rgb_int]

    @property
    def rgb_int(self) -> List[int]:
        """Returns class color in RGB representation

        :return: A color as [R, G, B] list of integers
        """

        hex_val = self.values[2][1:]
        return [int(hex_val[i : i + 2], 16) for i in (0, 2, 4)]

    @classmethod
    def has_value(cls, value: object) -> bool:
        """Checks if value is defined for one of the enum constants

        Example:
            LULC.has_value('forest')

        :param value: Any value
        :return: True if value is in enum and false otherwise
        """
        return value in cls._value2member_map_

    @classmethod
    def get_sentinel_hub_evaluation_function(cls, band_name: str) -> str:
        init = Template(
            "//VERSION=3 \n"
            "function setup(ds) { \n"
            "   setInputComponents([ds.$band_name]); \n"
            "   setOutputComponentCount(3); \n"
            "}\n"
            "\n"
            "function evaluatePixel(samples, scenes) {\n"
            "$script\n"
            "}\n"
        )

        script = []
        eval_pixel_script = Template("   if(samples[0].$band_name == $class_id) { return [$colors] }")
        for class_enum in cls:
            colors = ", ".join([f"{c}/255" for c in class_enum.rgb_int])
            script.append(eval_pixel_script.substitute(band_name=band_name, class_id=class_enum.id, colors=colors))

        return init.substitute(band_name=band_name, script="\n".join(script))

    @classmethod
    def get_sentinel_hub_legend(cls) -> str:
        legend = {"type": "discrete", "items": []}
        for class_enum in cls:
            legend["items"].append({"color": class_enum.color, "label": class_enum.name})

        return json.dumps(legend, indent=2)



"""
A module with useful utilities related to batch processing
"""
from typing import List


def read_timestamps_from_orbits(userdata: dict) -> List[str]:
    """Parses batch orbits payload to obtain a list of timestamp strings"""
    if "orbits" in userdata:
        userdata = userdata["orbits"]

    return [orbit["tiles"][0]["date"] for orbit in userdata]


def read_timestamps(userdata: dict) -> List[str]:
    """Parses timestamps from a userdata dictionary"""
    return userdata["timestamps"]



"""
Conversion of batch results to EOPatches
"""
from typing import Optional, List, Dict

import fs
from pydantic import Field

from eolearn.core import (
    FeatureType,
    EOWorkflow,
    EONode,
    SaveTask,
    OverwritePermission,
    MergeFeatureTask,
    RemoveFeature,
    MergeEOPatchesTask,
)
from eolearn.io import ImportFromTiff
from eolearn.features import LinearFunctionTask

from ..core.pipeline import Pipeline
from ..core.schemas import BaseSchema
from ..tasks.batch_to_eopatch import LoadUserDataTask, FixImportedTimeDependentFeatureTask, DeleteFilesTask


class FeatureMappingSchema(BaseSchema):
    """Defines a mapping between 1 or more batch outputs into an EOPatch feature"""

    batch_files: List[str] = Field(
        description=(
            "A list of files that will be converted into an EOPatch feature. If you specify multiple tiff "
            "files they will be concatenated together over the bands dimension in the specified order."
        ),
    )
    feature_type: FeatureType
    feature_name: str
    multiply_factor: Optional[float] = Field(description="Factor used to multiply feature values with.")


class BatchToEOPatchPipeline(Pipeline):
    class Schema(Pipeline.Schema):
        folder_key: str = Field(description="Storage manager key pointing to the path with Batch results")
        mapping: List[FeatureMappingSchema] = Field(
            description="A list of mapping from batch files into EOPatch features."
        )
        userdata_feature_name: Optional[str] = Field(
            description="A name of META_INFO feature in which userdata.json would be stored."
        )
        userdata_timestamp_reader: Optional[str] = Field(
            description=(
                "Either an import path to a utility function or a Python code describing how to read "
                "dates from userdata dictionary."
            ),
            example="\"[info['date'] for info in json.loads(userdata['metadata'])]\"",
        )

    def __init__(self, *args, **kwargs):
        """Additionally sets some basic parameters calculated from config parameters"""
        super().__init__(*args, **kwargs)

        self._processing_folder = self.storage.get_folder(self.config.folder_key, full_path=True)
        self._relative_processing_folder = self.storage.get_folder(self.config.folder_key)
        self._all_batch_files = self._get_all_batch_files()

    def filter_patch_list(self, patch_list: List[str]) -> List[str]:
        """Checks the content of EOPatches in the project folder. If some files, produced by Sentinel Hub Batch, still
        exist this means that EOPatch hasn't been yet successfully processed by this pipeline. That is because this
        pipeline deletes batch files at the end. Such patches will remain after this filtering.
        """
        processed_eopatches = set()
        for eopatch_name in patch_list:
            eopatch_folder = fs.path.combine(self._relative_processing_folder, eopatch_name)
            eopatch_files = set(self.storage.filesystem.listdir(eopatch_folder))

            if not any(batch_file in eopatch_files for batch_file in self._all_batch_files):
                processed_eopatches.add(eopatch_name)

        return [eopatch for eopatch in patch_list if eopatch not in processed_eopatches]

    def build_workflow(self) -> EOWorkflow:
        """Builds the workflow"""
        userdata_node = EONode(
            LoadUserDataTask(
                path=self._processing_folder,
                userdata_feature_name=self.config.userdata_feature_name,
                userdata_timestamp_reader=self.config.userdata_timestamp_reader,
                config=self.sh_config,
            )
        )

        mapping_nodes = [
            self._get_tiff_mapping_node(feature_mapping, userdata_node) for feature_mapping in self.config.mapping
        ]

        last_node = mapping_nodes[0] if len(mapping_nodes) == 1 else userdata_node
        if len(mapping_nodes) > 1:
            last_node = EONode(MergeEOPatchesTask(), inputs=mapping_nodes)

        processing_node = self.get_processing_node(last_node)

        save_task = SaveTask(
            path=self._processing_folder,
            compress_level=1,
            overwrite_permission=OverwritePermission.OVERWRITE_FEATURES,
            config=self.sh_config,
        )
        save_node = EONode(save_task, inputs=[processing_node])

        delete_task = DeleteFilesTask(
            path=self._processing_folder, filenames=self._all_batch_files, config=self.sh_config
        )
        cleanup_node = EONode(delete_task, inputs=[save_node], name="Delete batch data")

        return EOWorkflow.from_endnodes(cleanup_node)

    def _get_tiff_mapping_node(self, mapping: FeatureMappingSchema, previous_node: EONode) -> EONode:
        """Prepares tasks and dependencies that convert tiff files into an EOPatch feature"""
        if not all(batch_file.endswith(".tif") for batch_file in mapping.batch_files):
            raise ValueError(f"All batch files should end with .tif but found {mapping.batch_files}")

        feature_type = mapping.feature_type
        if not (feature_type.is_spatial() and feature_type.is_raster()):
            raise ValueError(f"Tiffs can only be read into spatial raster feature types, but {feature_type} was given.")

        tmp_features = []
        end_nodes = []
        for batch_file in mapping.batch_files:
            feature = feature_type, batch_file.replace(".tif", "_tmp")
            tmp_features.append(feature)

            tmp_timeless_feature = (
                FeatureType.MASK_TIMELESS if feature_type.is_discrete() else FeatureType.DATA_TIMELESS
            ), feature[1]

            import_task = ImportFromTiff(tmp_timeless_feature, folder=self._processing_folder, config=self.sh_config)
            # Filename is written into the dependency name to be used later for execution arguments:
            import_node = EONode(import_task, inputs=[previous_node], name=f"{batch_file} import")

            if feature_type.is_temporal():
                fix_task = FixImportedTimeDependentFeatureTask(tmp_timeless_feature, feature)
                end_nodes.append(EONode(fix_task, inputs=[import_node]))
            else:
                end_nodes.append(import_node)

        if len(end_nodes) == 1:
            previous_node = end_nodes[0]
        if len(end_nodes) > 1:
            previous_node = EONode(MergeEOPatchesTask(), inputs=end_nodes)

        final_feature = feature_type, mapping.feature_name
        merge_feature_task = MergeFeatureTask(input_features=tmp_features, output_feature=final_feature)
        merge_node = EONode(merge_feature_task, inputs=[previous_node])

        end_node = EONode(RemoveFeature(tmp_features), inputs=[merge_node])

        if mapping.multiply_factor is not None:
            multiply_task = LinearFunctionTask(final_feature, slope=mapping.multiply_factor)
            end_node = EONode(multiply_task, inputs=[end_node])

        return end_node

    @staticmethod
    def get_processing_node(previous_node: EONode) -> EONode:
        """This method can be overwritten to add more tasks that process loaded data before saving it."""
        return previous_node

    def get_execution_arguments(self, workflow: EOWorkflow) -> List[Dict[EONode, Dict[str, object]]]:
        """Prepare execution arguments per each EOPatch"""
        exec_args = super().get_execution_arguments(workflow)

        nodes = workflow.get_nodes()

        for name, single_exec_dict in zip(self.patch_list, exec_args):
            for node in nodes:
                if isinstance(node.task, ImportFromTiff):
                    filename = node.name.split()[0]
                    path = f"{name}/{filename}"
                    single_exec_dict[node] = dict(filename=path)

                if isinstance(node.task, (DeleteFilesTask, LoadUserDataTask)):
                    single_exec_dict[node] = dict(folder=name)

        return exec_args

    def _get_all_batch_files(self) -> List[str]:
        """Provides a list of batch files used in this pipeline"""
        files = [file for feature_mapping in self.config.mapping for file in feature_mapping.batch_files]

        if self.config.userdata_feature_name or self.config.userdata_timestamps_path:
            files.append("userdata.json")

        return list(set(files))



"""
Task definitions for prediction
"""
import abc
from typing import List, Optional, Tuple, Callable

import numpy as np
import joblib

from sentinelhub import SHConfig
from eolearn.core import EOPatch, EOTask, execute_with_mp_lock, get_filesystem

from ..utils.types import Feature


class BasePredictionTask(EOTask, metaclass=abc.ABCMeta):
    """Base predictions task streamlining data preprocessing before prediction"""

    def __init__(
        self,
        *,
        model_folder: str,
        model_filename: str,
        input_features: List[Feature],
        mask_feature: Feature,
        output_feature: Feature,
        mp_lock: bool,
        sh_config: SHConfig,
    ):
        """
        :param model_folder: Path to the folder where model is stored
        :param model_filename: Name of file containing the model
        :param input_features: List of features containing input for the model, which are concatenated in given order
        :param mask_feature: Mask specifying which points are to be predicted
        :param output_feature: Feature into which predictions are saved
        :param mp_lock: If predictions should be executed with a multiprocessing lock
        :param sh_config: SentinelHub config
        """

        self.model_folder = model_folder
        self.model_filename = model_filename
        self._model = None

        self.input_features = input_features
        self.mask_feature = mask_feature
        self.output_feature = output_feature

        self.mp_lock = mp_lock
        self.sh_config = sh_config

    def process_data(self, eopatch: EOPatch, mask: np.ndarray) -> List[np.ndarray]:
        """Masks and reshapes data into a form suitable for the model"""
        all_features = []
        for ftype, fname in self.input_features:
            array = eopatch[ftype, fname]

            if ftype.is_timeless():
                all_features.append(array[mask, :])
            else:
                array = array[:, mask, :]
                time, pixels, depth = array.shape
                array = np.moveaxis(array, 0, 1)
                all_features.append(array.reshape(pixels, time * depth))

        return np.concatenate(all_features, axis=-1)

    @property
    def model(self):
        """Implements lazy loading that gets around file-system issues"""
        if self._model is None:
            file_system = get_filesystem(self.model_folder, config=self.sh_config)
            with file_system.openbin(self.model_filename, "r") as file_handle:
                self._model = joblib.load(file_handle)
        return self._model

    def apply_predictor(
        self, predictor: Callable, processed_features: np.ndarray, return_on_empty: Optional[np.ndarray] = None
    ) -> np.ndarray:
        """Helper function that applies the predictor according to the mp_lock settings"""
        if processed_features.shape[0] == 0 and return_on_empty is not None:
            return return_on_empty

        if self.mp_lock:
            return execute_with_mp_lock(predictor, processed_features)
        return predictor(processed_features)

    @abc.abstractmethod
    def add_predictions(self, eopatch: EOPatch, processed_features: np.ndarray, mask: np.ndarray) -> EOPatch:
        """Runs the model prediction on given features and adds them to the eopatch. Must reverse mask beforehand."""

    @staticmethod
    def transform_to_feature_form(predictions: np.ndarray, mask: np.ndarray, no_value=0) -> np.ndarray:
        """Transforms an array of predictions into an EOPatch suitable array, making sure to reverse the masking"""
        full_predictions = np.full((*mask.shape, predictions.shape[-1]), dtype=predictions.dtype, fill_value=no_value)
        full_predictions[mask, :] = predictions
        return full_predictions

    def execute(self, eopatch: EOPatch) -> EOPatch:
        """Run model on input features and save predictions to eopatch"""

        some_feature = self.input_features[0]
        mask_size = eopatch.get_spatial_dimension(*some_feature)
        mask = np.squeeze(eopatch[self.mask_feature], axis=-1) if self.mask_feature else np.ones(mask_size)
        mask = mask.astype(bool)

        preprocessed_features = self.process_data(eopatch, mask)
        eopatch = self.add_predictions(eopatch, preprocessed_features, mask)
        return eopatch


class ClassificationPredictionTask(BasePredictionTask):
    """Uses a classification model to produce predictions for given input features"""

    def __init__(
        self,
        *,
        label_encoder_filename: str,
        output_probability_feature: Optional[Feature] = None,
        **kwargs,
    ):
        """
        :param label_encoder_filename: Name of file containing the label encoder with which to decode predictions
        :param output_probability_feature: If specified saves pseudo-probabilities into given feature.
        :param kwargs: Parameters of `BasePredictionTask`
        """
        self.label_encoder_filename = label_encoder_filename
        self._label_encoder = None
        self.output_probability_feature = output_probability_feature
        super().__init__(**kwargs)

    @property
    def label_encoder(self):
        """Implements lazy loading that gets around file-system issues"""
        if self._label_encoder is None and self.label_encoder_filename is not None:
            file_system = get_filesystem(self.model_folder, config=self.sh_config)
            with file_system.openbin(self.label_encoder_filename, "r") as file_handle:
                self._label_encoder = joblib.load(file_handle)
        return self._label_encoder

    def add_predictions(self, eopatch: EOPatch, processed_features: np.ndarray, mask: np.ndarray) -> EOPatch:
        """Runs the model prediction on given features and adds them to the eopatch

        If specified also adds probability scores and uses a label encoder.
        """

        predictions = self.apply_predictor(self.model.predict, processed_features, np.zeros((0,), dtype=np.uint8))
        predictions = predictions[..., np.newaxis]
        if self.label_encoder is not None:
            predictions = self.label_encoder.inverse_transform(predictions)
        eopatch[self.output_feature] = self.transform_to_feature_form(predictions, mask)

        if self.output_probability_feature is not None:
            probabilities = self.apply_predictor(self.model.predict_proba, processed_features)
            eopatch[self.output_probability_feature] = self.transform_to_feature_form(probabilities, mask)

        return eopatch


class RegressionPredictionTask(BasePredictionTask):
    """Computes values and scores given an input model and eopatch feature name"""

    def __init__(
        self,
        *,
        clip_predictions: Optional[Tuple[float, float]],
        **kwargs,
    ):
        """
        :param clip_predictions: If given the task also clips predictions to the specified interval.
        :param kwargs: Parameters of `BasePredictionTask`
        """
        self.clip_predictions = clip_predictions
        super().__init__(**kwargs)

    def add_predictions(self, eopatch: EOPatch, processed_features: np.ndarray, mask: np.ndarray) -> EOPatch:
        """Runs the model prediction on given features and adds them to the eopatch. Must reverse mask beforehand."""

        predictions = self.apply_predictor(self.model.predict, processed_features, np.zeros((0,), dtype=np.float32))
        predictions = predictions[..., np.newaxis]
        if self.clip_predictions is not None:
            predictions = predictions.clip(*self.clip_predictions)
        eopatch[self.output_feature] = self.transform_to_feature_form(predictions, mask)

        return eopatch



"""
Utilities used for logging
"""
from json.decoder import JSONDecodeError

import requests

INSTANCE_INFO_URL = "http://169.254.169.254/latest/dynamic/instance-identity/document"
INSTANCE_REQUEST_TIMEOUT = 0.1


def get_instance_info() -> dict:
    """Provides information about a compute instance on which this code is being executed

    For now this method is designed only to collect info about AWS instances. For more info check:
    https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/instance-identity-documents.html

    :return: A dictionary with information
    :rtype: dict
    """
    try:
        response = requests.get(INSTANCE_INFO_URL, timeout=INSTANCE_REQUEST_TIMEOUT)
        response.raise_for_status()
        return response.json()
    except requests.RequestException:
        return {"error": "Not an AWS instance or AWS IMDSv1 service not available"}
    except JSONDecodeError:
        return {"error": f"Failed to decode a response from {INSTANCE_INFO_URL}"}
